#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Nov 29, 2023
email    : Nasy <nasyxx+python@gmail.com>
filename : metrics.py
project  : nadl
license  : GPL-3.0+

* Metric class and functions
* Some simple metrics that not in optax loss.  For complex metrics, use sklearn.
"""

# ruff: noqa: F722
from functools import partial

import jax
import jax.numpy as jnp
from equinox import (
  AbstractVar,
  Module,
  combine,
  filter_pure_callback,
  is_array,
  partition,
  tree_equal,
  tree_pformat,
)

from jaxtyping import Array, ArrayLike, Int, Num
from typing import Literal, Self

import sklearn.metrics as m

from .utils import all_array, filter_concat

type N = Num[Array, "#a"]


def convert(x: ArrayLike | N) -> N:
  """Convert to float."""
  if isinstance(x, jax.Array):
    return x.reshape(-1, *x.shape[1:])
  return jnp.asarray(x).reshape(-1)


def roc_auc_score(labels: Num[Array, " A"], preds: Num[Array, " A"]) -> N:
  """Compute ROC."""
  return jax.pure_callback(
    m.roc_auc_score, jax.ShapeDtypeStruct((), jnp.float32), labels, preds
  )


def average_precision_score(
  labels: Num[Array, " A"],
  preds: Num[Array, " A"],
  average: Literal["micro", "macro"] = "macro",
) -> N:
  """Compute PR."""
  return jax.pure_callback(
    partial(m.average_precision_score, average=average),
    jax.ShapeDtypeStruct((), jnp.float32),
    labels,
    preds,
  )


def pr_auc_score(labels: Num[Array, " A"], preds: Num[Array, " A"]) -> N:
  """Compute PR."""

  def _callback(lbl: Num[Array, " A"], prd: Num[Array, " A"]) -> N:
    precision, recall, _ = m.precision_recall_curve(lbl, prd)
    return jnp.asarray(m.auc(recall, precision))

  return filter_pure_callback(
    _callback,
    labels,
    preds,
    result_shape_dtypes=jax.ShapeDtypeStruct((), jnp.float32),
  )


class Metric[**P, T](Module):
  """Base Metric."""

  name: AbstractVar[str | None]

  def __post_init__(self) -> None:
    """Post init."""
    for k, v in self.__dict__.items():
      if isinstance(v, Array):
        self.__dict__[k] = convert(v)

  def __check_init__(self) -> None:  # noqa: PLW3201
    """Check init."""
    arrs = all_array(self)
    if arrs and (not tree_equal(*jax.tree.map(lambda x: jnp.shape(x)[0], arrs))):
      raise ValueError("All metrics should have the same length in first dim.")

  @classmethod
  def merge(cls, *metrics: Self) -> Self:
    """Merge all metrics."""
    _, s2 = partition(metrics, is_array)
    if not tree_equal(*s2):
      raise ValueError("All metrics should have the same non-array values.")
    return filter_concat(metrics)

  def __add__(self, value: Self) -> Self:
    """Add."""
    return self.merge(self, value)

  def __or__(self, value: Self) -> Self:
    """Or."""
    return combine(self, value)

  @classmethod
  def empty(cls, *args: P.args, **kwds: P.kwargs) -> Self:  # noqa: ARG003
    """Empty."""
    return cls(dict.fromkeys(cls.__dataclass_fields__))  # type: ignore

  def compute(self) -> T:
    """Compute."""
    raise NotImplementedError

  def __getitem__(self, idx: int | slice | Int[ArrayLike, "..."]) -> Self:
    """Get item."""
    return jax.tree.map(
      lambda x: x[idx] if is_array(x) else x,
      self,
    )

  def show(self) -> str:
    """Show."""
    return tree_pformat(self, short_arrays=False)


class Accuracy(Metric):
  """Accuracy."""

  labels: Int[Array, "..."]
  preds: Int[Array, "..."]
  name: str = "accuracy"

  @classmethod
  def empty(cls, name: str = "accuracy") -> Self:
    """Empty."""
    return cls(labels=jnp.asarray([jnp.nan]), preds=jnp.asarray([jnp.nan]), name=name)

  def compute(self) -> Array:
    """Compute."""
    return jnp.nanmean(self.labels == self.preds, axis=-1)


def dice_coef(y_true: jax.Array, y_pred: jax.Array, eps: float = 1e-8) -> jax.Array:
  """Compute dice coefficient."""
  y_true = jnp.asarray(y_true)
  y_pred = jnp.asarray(y_pred)

  intersection = jnp.sum(y_true * y_pred)
  union = jnp.sum(y_true) + jnp.sum(y_pred)

  return (2.0 * intersection) / (union + eps)


def iou_coef(y_true: jax.Array, y_pred: jax.Array, eps: float = 1e-8) -> jax.Array:
  """Compute intersection over union."""
  y_true = jnp.asarray(y_true)
  y_pred = jnp.asarray(y_pred)

  intersection = jnp.sum(y_true * y_pred)
  union = jnp.sum(y_true) + jnp.sum(y_pred) - intersection

  return intersection / (union + eps)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : May 14, 2024
email    : Nasy <nasyxx+python@gmail.com>
filename : states.py
project  : nadl
license  : GPL-3.0+

States
"""

from __future__ import annotations

import shutil
from abc import abstractmethod

import equinox as eqx
import optax
import orbax.checkpoint as ocp

from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
  from collections.abc import Callable
  from pathlib import Path

  import jax

  from jaxtyping import PyTree

  from rich.console import Console

  from .metrics import Metric


class BaseTrainState[T, M](eqx.Module):
  """Train state."""

  model: M
  tx: optax.GradientTransformation
  opt_state: optax.OptState
  loss: jax.Array
  step: jax.Array
  conf: T

  @classmethod
  @abstractmethod
  def create[**P](
    cls: type[BaseTrainState[T, M]], *args: P.args, **kwds: P.kwargs
  ) -> BaseTrainState[T, M]:
    """Create state."""
    raise NotImplementedError

  def apply_grads(self, loss: jax.Array, grads: eqx.Module) -> BaseTrainState[T, M]:
    """Apply gradients."""
    updates, opt_state = self.tx.update(
      cast(optax.Updates, grads), self.opt_state, params=cast(optax.Params, self.model)
    )
    model = eqx.apply_updates(self.model, updates)
    return eqx.tree_at(
      lambda x: (x.model, x.opt_state, x.loss, x.step),
      self,
      (model, opt_state, loss, self.step + 1),
    )


def state_fn(
  rpath: Path,
  console: Console | None = None,
  keeps: int = 5,
  clean: bool = False,
  item_names: tuple[str, ...] | None = None,
  item_handlers: dict | None = None,
  best_fn: Callable[[PyTree], float] | None = None,
) -> tuple[
  ocp.CheckpointManager,
  Callable[[int, BaseTrainState, Metric | None, dict[str, float] | None], None],
]:
  """Get states manager."""
  match (item_names, item_handlers):
    case (None, None):
      item_names = ("state", "extra_metadata")
      item_handlers = {
        "state": ocp.PyTreeCheckpointHandler(),
        "extra_metadata": ocp.PyTreeCheckpointHandler(),
      }
    case _ if item_names and item_handlers:
      for i in item_names:
        assert i in item_handlers, f"Item {i} not in item_handlers."
    case _:
      raise ValueError("item_names and item_handlers should be both None or not None.")

  if console:
    console.log(f"Checkpoint path at {rpath}")
  if rpath.exists() and clean:
    if console:
      console.log("Cleaning up checkpoint...")
    shutil.rmtree(rpath)

  mngr = ocp.CheckpointManager(
    rpath,
    options=ocp.CheckpointManagerOptions(
      max_to_keep=keeps, save_interval_steps=1, best_fn=best_fn
    ),
    item_names=item_names,
    item_handlers=item_handlers,
  )

  def save(
    step: int,
    state: BaseTrainState,
    metadata: Metric | None = None,
    metrics: dict[str, float] | None = None,
  ) -> None:
    """Save state."""
    mngr.save(
      step,
      args=ocp.args.Composite(
        state=ocp.args.PyTreeSave(state),  # type: ignore
        extra_metadata=ocp.args.PyTreeSave(metadata or {}),  # type: ignore
      ),
      metrics=metrics or {},
    )

  return mngr, save



#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Nov 30, 2023
email    : Nasy <nasyxx+python@gmail.com>
filename : loops.py
project  : nadl
license  : GPL-3.0+

Train Eval Loops
"""

from __future__ import annotations

from threading import Event, Thread

from equinox import Module

from typing import TYPE_CHECKING, Any, Self

from rich.console import Console
from rich.progress import (
  BarColumn,
  Progress,
  ProgressColumn,
  TaskID,
  TextColumn,
  TimeElapsedColumn,
  TimeRemainingColumn,
)
from rich.theme import Theme

from .utils import pformat

if TYPE_CHECKING:
  from collections.abc import Hashable, Mapping

  from types import TracebackType

DEF_LIGHT_THEME = Theme({
  "bar.back": "#50616D",
  "bar.complete": "#D3CBAF",
  "bar.finished": "#b49b7f",
})


RESC = TextColumn("{task.fields[res]}")


class PGThread(Thread):
  """Progress thread."""

  def __init__(self, pg: Progress, task_id: TaskID) -> None:
    """Init."""
    super().__init__()
    self.pg = pg
    self.tid = task_id
    self.done = Event()
    self.completed = 0
    self.res: str | None = None
    super().__init__()

  def run(self) -> None:
    """Run."""
    last_completed = 0
    wait = self.done.wait
    while not wait(0.2):
      if (completed := self.completed) != last_completed:
        self.pg.advance(self.tid, completed - last_completed)
        last_completed = completed
        if self.res is not None:
          self.pg.update(self.tid, res=self.res)
          self.res = None
    self.pg.update(self.tid, completed=self.completed, refresh=True)

  def __enter__(self) -> Self:
    """Enter."""
    self.start()
    return self

  def __exit__(
    self,
    exc_type: type[BaseException] | None,
    exc_val: BaseException | None,
    exc_tb: TracebackType | None,
  ) -> None:
    """Exit."""
    del exc_type, exc_val, exc_tb
    self.done.set()
    self.join()


class PG(Module):
  """Progress."""

  pg: Progress
  console: Console
  tasks: dict[Hashable, TaskID]

  @classmethod
  def init_progress(
    cls: type[Self],
    pg: Progress | None = None,
    console: Console | None = None,
    total: bool = True,
    bar_width: int | None = 20,
    extra_columns: tuple[ProgressColumn, ...] = (),
    show_progress: bool = True,
    theme: Theme | None = None,
  ) -> Self:
    """Init progress bar."""
    if console is None:
      console = Console(theme=theme or DEF_LIGHT_THEME)
    if pg is None:
      pg = Progress(
        TextColumn(
          "{task.description}" + " - {task.completed}/{task.total}" if total else ""
        ),
        TimeRemainingColumn(),
        TimeElapsedColumn(),
        BarColumn(bar_width),
        console=console,
        disable=not show_progress,
      )
    pg.columns += extra_columns
    return cls(pg, console, {})

  def add_columns(self, columns: tuple[ProgressColumn, ...]) -> Self:
    """Add columns."""
    self.pg.columns += columns
    return self

  def add_task(
    self,
    description: str,
    start: bool = True,
    total: float | None = 100,
    completed: int = 0,
    visible: bool = True,
    **fileds: Any,  # noqa: ANN401
  ) -> TaskID:
    """Add task."""
    task = self.pg.add_task(
      description,
      start=start,
      total=total,
      completed=completed,
      visible=visible,
      **fileds,
    )
    self.tasks[description] = task
    return task

  def advance(self, task: TaskID, advance: float = 1) -> None:
    """Advance task."""
    self.pg.advance(task, advance=advance)

  def __enter__(self) -> Progress:
    """Enter."""
    return self.pg.__enter__()

  def __exit__(
    self,
    exc_type: type[BaseException] | None,
    exc_val: BaseException | None,
    exc_tb: TracebackType | None,
  ) -> None:
    """Exit."""
    self.pg.__exit__(exc_type, exc_val, exc_tb)

  def update_res(
    self, name: str, updates: Mapping[str, float | int | str | None]
  ) -> None:
    """Update res."""
    if name in self.tasks:
      self.pg.update(self.tasks[name], res=pformat(updates))


def test() -> None:
  """Test progress."""
  pg = PG.init_progress()
  t0 = pg.add_task("Task 0", total=10**8)
  with pg, PGThread(pg.pg, t0) as pt:
    for _i in range(10**8):
      pt.completed += 1


if __name__ == "__main__":
  test()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Apr 26, 2024
email    : Nasy <nasyxx+python@gmail.com>
filename : images.py
project  : nadl
license  : GPL-3.0+

Images related utilities.
"""

import jax
import jax.numpy as jnp
from jaxtyping import Float, Array, Integer


def make_patches(
  image: Float[Array, "h w c"] | Integer[Array, "h w c"],
  patch_sizes: tuple[int, int],
  stride_sizes: tuple[int, int],
) -> jax.Array:
  """Extract patches from an input image."""
  image = jnp.asarray(image)

  # Unpack patch and stride sizes
  patch_height, patch_width = patch_sizes
  stride_height, stride_width = stride_sizes

  # Image dimensions
  img_height, img_width, _ = image.shape

  # Compute number of patches along each dimension
  num_patches_height = (img_height - patch_height) // stride_height + 1
  num_patches_width = (img_width - patch_width) // stride_width + 1

  # Creating patch index grid
  i, j = jnp.meshgrid(
    jnp.arange(num_patches_height) * stride_height,
    jnp.arange(num_patches_width) * stride_width,
    indexing="ij",
  )

  # Extract patches using advanced indexing
  return image[
    i[:, :, None, None] + jnp.arange(patch_height)[None, None, :, None],
    j[:, :, None, None] + jnp.arange(patch_width)[None, None, None, :],
    None,
  ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Nov 29, 2023
email    : Nasy <nasyxx+python@gmail.com>
filename : __init__.py
project  : nadl
license  : GPL-3.0+

NADL
"""

from .data import DState, IdxDataloader, es_loop
from .keys import Keys, new_key
from .loops import PG, RESC, PGThread
from .metrics import (
  Accuracy,
  Metric,
  average_precision_score,
  convert,
  dice_coef,
  iou_coef,
  pr_auc_score,
  roc_auc_score,
)
from .preprocessing import (
  SCALER,
  identity_scaler,
  min_max_scaler,
  normalizer,
  select_scaler,
  standard_scaler,
)
from .states import BaseTrainState, state_fn
from .utils import (
  all_array,
  classit,
  filter_concat,
  pformat,
  rle,
  rle_array,
)

__version__ = "1.5.9"

__all__ = [
  "PG",
  "RESC",
  "SCALER",
  "Accuracy",
  "BaseTrainState",
  "DState",
  "IdxDataloader",
  "Keys",
  "Metric",
  "Metric",
  "PGThread",
  "all_array",
  "average_precision_score",
  "classit",
  "convert",
  "dice_coef",
  "es_loop",
  "filter_concat",
  "identity_scaler",
  "iou_coef",
  "min_max_scaler",
  "new_key",
  "normalizer",
  "pformat",
  "pr_auc_score",
  "rle",
  "rle_array",
  "roc_auc_score",
  "select_scaler",
  "standard_scaler",
  "state_fn",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Nov 29, 2023
email    : Nasy <nasyxx+python@gmail.com>
filename : keys.py
project  : nadl
license  : GPL-3.0+

Keys.
"""

from __future__ import annotations

import jax
import jax.numpy as jnp
from equinox import Module, field, tree_at

from typing import TYPE_CHECKING

if TYPE_CHECKING:
  from collections.abc import Sequence


class Keys(Module):
  """JAX random key sequence.

  init_key:  The initial key.
  keys: The key sequence in cluding the histories.
  idx: Current key idx in the key sequence.
  """

  init_key: jax.Array
  keys: list[jax.Array]
  idx: jax.Array = field(default_factory=lambda: jax.numpy.array(0))

  @property
  def key(self) -> jax.Array:
    """Get key."""
    if self.keys:
      return self.keys[-1]
    self.reserve(1)
    return self.keys[-1]

  @property
  def state(self) -> tuple[jax.Array, Sequence[jax.Array], jax.Array]:
    """Get state."""
    return self.init_key, self.keys, self.idx

  @classmethod
  def from_int_or_key(cls: type[Keys], key: jax.Array | int) -> Keys:
    """Convert int or key to Keys."""
    if isinstance(key, int):
      return cls(jax.random.key(key), [])
    return cls(key, [])

  @classmethod
  def from_state(
    cls: type[Keys], key: jax.Array, keys: Sequence[jax.Array], idx: jax.Array
  ) -> Keys:
    """Convert state to Keys."""
    return cls(key, list(keys), idx)

  def reserve(self, num: int | jax.Array) -> Keys:
    """Reverse the keys."""
    while (num := num - 1) >= 0:
      if self.keys:
        self.keys.append(jax.random.fold_in(self.keys[-1], 0))
      else:
        self.keys.append(jax.random.fold_in(self.init_key, 0))
    return self

  def next_key(self) -> Keys:
    """Get next key."""
    return tree_at(lambda k: k.idx, self, self.idx + 1)

  def __iter__(self) -> Keys:
    """Iterate the keys."""
    return self

  def __next__(self) -> Keys:
    """Get next key."""
    return self.next_key()

  def take(self, num: int) -> jax.Array:
    """Take num keys."""
    self.reserve(jnp.max(num - len(self.keys), 0))
    return jnp.r_[*self.keys[-num:]]

  def __len__(self) -> int:
    """Length of keys."""
    return len(self.keys)

  def __call__(self, epoch: int | jax.Array | None = None) -> jax.Array:
    """Get keys for epoch."""
    match epoch:
      case int():
        if epoch + 1 > len(self.keys):
          self.reserve(epoch + 1 - len(self.keys))
        return self.keys[epoch]
      case jax.Array():
        self.reserve(jnp.maximum((epoch + 1 - len(self.keys)).max(), 0))
        return jnp.r_[*self.keys][epoch]
      case None:
        return self(self.idx)


new_key = Keys.from_int_or_key
